/*
 * condump.c - Windows console color snapshot tool.
 *
 * Jason Hood, 20 to 22 December, 2012.
 *
 * Based directly on:
 *
 * vcsadump - Linux virtual console color snapshot tool
 *
 * (C) 2003 Florian <petafrog@yahoo.com>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#define PVERS "1.00"
#define PDATE "22 December, 2012"


#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include "getopt.h"


#define PROGNAME "condump"
#define CSSNAME  "attrib.css"
#define HTMLNAME "dump.html"
#define ANSINAME "dump.ans"


CONSOLE_SCREEN_BUFFER_INFO csbi;
int  skip = 2, lines;
BOOL window;


// Should possibly read these from the registry (HKCU\Console\ColorTableNN).
static const char* colormap[] =
{
  "#000000", "#000080", "#008000", "#008080",
  "#800000", "#800080", "#808000", "#C0C0C0",
  "#808080", "#0000FF", "#00FF00", "#00FFFF",
  "#FF0000", "#FF00FF", "#FFFF00", "#FFFFFF"
};


// Map control characters using the CP437 glyphs as Unicode codepoints.
static const WCHAR ctrlmap[32] =
{
  ' ',       //  0
  L'\x263A', //  1 White Smiling Face
  L'\x263B', //  2 Black Smiling Face
  L'\x2665', //  3 Black Heart Suit
  L'\x2666', //  4 Black Diamond Suit
  L'\x2663', //  5 Black Club Suit
  L'\x2660', //  6 Black Spade Suit
  L'\x2022', //  7 Bullet
  L'\x25D8', //  8 Inverse Bullet
  L'\x25CB', //  9 White Circle
  L'\x25D9', // 10 Inverse White Circle
  L'\x2642', // 11 Male Sign
  L'\x2640', // 12 Female Sign
  L'\x266A', // 13 Eigth Note
  L'\x266B', // 14 Beamed Eigth Notes
  L'\x263C', // 15 White Sun With Rays
  L'\x25BA', // 16 Black Right-Pointing Pointer
  L'\x25C4', // 17 Black Left-Pointing Pointer
  L'\x2195', // 18 Up Down Arrow
  L'\x203C', // 19 Double Exclamation Mark
  L'\x00B6', // 20 Pilcrow Sign
  L'\x00A7', // 21 Section Sign
  L'\x25AC', // 22 Black Rectangle
  L'\x21A8', // 23 Up Down Arrow With Base
  L'\x2191', // 24 Upwards Arrow
  L'\x2193', // 25 Downwards Arrow
  L'\x2192', // 26 Rightwards Arrow
  L'\x2190', // 27 Leftwards Arrow
  L'\x221F', // 28 Right Angle
  L'\x2194', // 29 Left Right Arrow
  L'\x25B2', // 30 Black Up-Pointing Pointer
  L'\x25BC'  // 31 Black Down-Pointing Pointer
};


/**
 * Dumps usage information to stdout and exits.
 */
void usage( void )
{
  printf(
"Console Dump by Jason Hood <jadoxa@yahoo.com.au>.\n"
"Version " PVERS " (" PDATE ").  GPL.\n"
"http://condump.adoxa.cjb.net/\n"
"\n"
"%s [-abCdehkpuUw] [-c fn] [-l N] [-o fn] [-s N] [-t s]\n"
"\n"
"Use exactly one of the following:\n"
"-h\twrite HTML page with inline attributes\n"
"-e\twrite HTML page with link to external CSS\n"
"-p\twrite HTML fragment containing screen dump only\n"
"-C\twrite external CSS containing all possible attributes (no dump made)\n"
"-a\twrite ANSI file (current code page)\n"
"-u\twrite ANSI file (UTF-8)\n"
"-U\twrite ANSI file (UTF-8 with BOM)\n"
"\n"
"Use zero or more of the following:\n"
"-d\tHTML: write colors directly in span tag (no style sheet)\n"
"\tANSI: write trailing spaces directly (rather than erasing the line)\n"
"-s N\tskip the previous N lines (default is %d)\n"
"-b\tdump the buffer rather than the window\n"
"-w\tdump the entire window (ignores -s, -l and current cursor position)\n"
"-l N\tdump the previous N lines (after the skip)\n"
"-k\twait for a key before dumping (use the mouse to scroll the window)\n"
"-t s\tset HTML title (default is current console title)\n"
"-o fn\tset output filename (default \"%s\" or \"%s\")\n"
"-c fn\tset external CSS filename (default \"%s\")\n"
"\n"
"Using \"-\" as a filename causes output to be written to stdout.\n"
	 , PROGNAME, skip, HTMLNAME, ANSINAME, CSSNAME );
  exit( 0 );
}


/**
 * Dumps a style sheet containing console attributes to the given
 * file pointer.  If attribs_used is non-null, then it must point to an array
 * of 256 chars, and only those attributes for which the corresponding char
 * is nonzero will be used.
 */
void dumpcss( FILE* fp, char* attribs_used )
{
  int i;

  // If making full CSS, print a small header
  if (!attribs_used)
    fputs( "/* Windows console attributes automatically generated by " PROGNAME " */\n\n", fp );

  for (i = 0; i < 256; i++)
  {
    if (!attribs_used || attribs_used[i])
      fprintf( fp, "span.a%02X { color: %s; background-color: %s; }\n",
	       i, colormap[i & 0x0F], colormap[(i >> 4) & 0x0F] );
  }
}


/**
 * Scans the dump and marks which attributes are used.
 */
void scanattribs( LPWORD dump, char* attribs_used )
{
  int n = lines * csbi.dwSize.X;

  while (n--)
    attribs_used[*dump & 255] = TRUE, dump++;
}


/**
 * Encodes the given dump, writing HTML to the given file pointer.
 */
void htmlencode( FILE* fp, LPWSTR cdump, LPWORD adump, BOOL direct )
{
  int y, x, oldattr = -1;

  fputs( "<pre>", fp );
  for (y = 0; y < lines; y++)
  {
    for (x = 0; x < csbi.dwSize.X; x++)
    {
      WCHAR ch = *cdump++;
      int attr = *adump++;

      // If attribute changed, print appropriate tag.
      if (oldattr != attr)
      {
	if (oldattr != -1)
	  fputs( "</span>", fp ); // close old tag if open
	if (direct)
	  fprintf( fp, "<span style=\"color: %s; background: %s;\">",
		   colormap[attr & 0x0F], colormap[(attr >> 4) & 0x0F] );
	else
	  fprintf( fp, "<span class=\"a%02X\">", attr );
	oldattr = attr;
      }

      if (ch < 0x20)
	ch = ctrlmap[ch];
      else if (ch == 127)
	ch = L'\x2302'; // House

      // escape some HTML entities
      switch (ch)
      {
	case '<': fputs( "&lt;", fp ); break;
	case '>': fputs( "&gt;", fp ); break;
	case '&': fputs( "&amp;", fp ); break;

	default:
	 if (ch < 128)
	   putc( ch, fp );
	 else
	 {
	   char utf[8];
	   int size = WideCharToMultiByte( CP_UTF8, 0, &ch, 1, utf, sizeof(utf),
					   NULL, NULL );
	   fprintf( fp, "%.*s", size, utf );
	 }
      }
    }
    fputs( "\n", fp );
  }

  fputs( "</span></pre>\n", fp );
}


/**
 * Translate a Windows attribute to a minimal ANSI escape sequence.
 */
void translate_attr( FILE* fp, int attr )
{
  static int fg = -1, bg = -1, bold, underline;
  static const int attr2ansi[] = { 0, 4, 2, 6, 1, 5, 3, 7 };
  int  f, b, B, u;
  char seq[32];
  int  len;

  f = attr2ansi[attr & 7];
  b = attr2ansi[(attr >> 4) & 7];
  B = attr & 8;
  u = attr & 0x80;

  // Explicitly set the initial color.
  if (fg == -1)
  {
    bold = !B;
    underline = !u;
  }

  len = 0;
  seq[len++] = '\33';
  seq[len++] = '[';
  if (B && !bold)
  {
    seq[len++] = '1';
  }
  else if (!B && bold)
  {
    seq[len++] = '2';
    seq[len++] = '2';
  }
  if (f != fg)
  {
    if (B != bold)
      seq[len++] = ';';
    seq[len++] = '3';
    seq[len++] = '0' + f;
  }
  if (u && !underline)
  {
    if (len != 2)
      seq[len++] = ';';
    seq[len++] = '4';
  }
  else if (!u && underline)
  {
    if (len != 2)
      seq[len++] = ';';
    seq[len++] = '2';
    seq[len++] = '4';
  }
  if (b != bg)
  {
    if (u != underline)
      seq[len++] = ';';
    seq[len++] = '4';
    seq[len++] = '0' + b;
  }
  if (len == 2)
    return;

  seq[len++] = 'm';
  fwrite( seq, len, 1, fp );

  fg = f;
  bg = b;
  bold = B;
  underline = u;
}


/**
 * Encodes the given dump, writing ANSI to the given file pointer.
 */
void ansiencode( FILE* fp, LPWSTR cdump, LPWORD adump, UINT codepage, BOOL direct )
{
  int y, x, e, attr, oldattr = -1;
  WCHAR ch;

  for (y = 0; y < lines; y++)
  {
    e = csbi.dwSize.X;
    if (!direct)
    {
      // Translate trailing spaces to an erase line sequence.
      while (--e >= 0 && cdump[e] == ' ' && adump[e] == adump[csbi.dwSize.X-1])
	;
      ++e;
    }
    for (x = 0; x < e; x++)
    {
      ch   = cdump[x];
      attr = adump[x];

      // If attribute changed, print appropriate sequence.
      if (oldattr != attr)
      {
	translate_attr( fp, attr );
	oldattr = attr;
      }

      if (codepage == CP_UTF8)
      {
	if (ch < 0x20)
	  ch = ctrlmap[ch];
	else if (ch == 127)
	  ch = L'\x2302'; // House
      }
      if (ch < 128)
	putc( ch, fp );
      else
      {
	char mb[8];
	int size = WideCharToMultiByte( codepage, 0, &ch, 1, mb, sizeof(mb),
					NULL, NULL );
	if (size == 1)
	  putc( *mb, fp );
	else
	  fprintf( fp, "%.*s", size, mb );
      }
    }
    // Even if there's only one space, erase the line to avoid a probable
    // scroll when it's displayed.
    if (e != csbi.dwSize.X)
    {
      attr = adump[csbi.dwSize.X-1];
      if (oldattr != attr)
      {
	translate_attr( fp, attr );
	oldattr = attr;
      }
      fputs( "\33[K", fp );
    }
    fputs( "\n", fp );

    cdump += csbi.dwSize.X;
    adump += csbi.dwSize.X;
  }

  fputs( "\33[m", fp );
}


/**
 * Reads a character & attribute dump from the console. Returns nonzero in case
 * of errors.
 */
int readcon( LPWSTR* cdump, LPWORD* adump )
{
  int s;
  COORD pos;
  DWORD read;
  HANDLE con = CreateFileW( L"CONOUT$", GENERIC_READ | GENERIC_WRITE,
			    FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL );

  if (!GetConsoleScreenBufferInfo( con, &csbi ))
    return 1;

  pos.X = 0;
  if (window)
  {
    pos.Y = csbi.srWindow.Top;
    lines = csbi.srWindow.Bottom - csbi.srWindow.Top + 1;
  }
  else
  {
    pos.Y = csbi.dwCursorPosition.Y - skip;
    if (pos.Y < 0)
      pos.Y = 0;
    if (lines == 0)
    {
      // take the size of the window up to the command
      lines = pos.Y - csbi.srWindow.Top;
      // include the line that scrolled off because of the command
      if (csbi.dwCursorPosition.Y == csbi.srWindow.Bottom)
	++lines;
    }
    if (pos.Y - lines < 0)
      lines = pos.Y;
    pos.Y -= lines;
  }

  // allocate memory for console data
  s = lines * csbi.dwSize.X;
  if (!(*cdump = malloc( s * (sizeof(WCHAR) + sizeof(WORD)) )))
  {
    fprintf( stderr, "Can't allocate %d bytes for console dump buffer.\n",
	     s * (sizeof(WCHAR) + sizeof(WORD)) );
    exit( 4 );
  }
  *adump = (LPWORD)(*cdump + s);

  ReadConsoleOutputCharacterW( con, *cdump, s, pos, &read );
  ReadConsoleOutputAttribute(  con, *adump, s, pos, &read );

  CloseHandle( con );
  return 0;
}


/**
 * Prints first half of a simple HTML page to file pointer.
 */
void htmlheaders( FILE* fp, char* title, char* cssfn, char* attribs_used )
{
  fprintf( fp,
	   "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">\n"
	   "<html>\n"
	   "<head>\n"
	   "<meta HTTP-EQUIV=\"content-type\" CONTENT=\"text/html; charset=UTF-8\">\n"
	   "<title>%s</title>\n", title );
  if (cssfn)
    fprintf( fp, "<link type=\"text/css\" rel=\"stylesheet\" href=\"%s\">\n", cssfn );
  else if (attribs_used)
  {
    // embed attributes
    fputs( "<style type=\"text/css\">\n", fp );
    dumpcss( fp, attribs_used );
    fputs( "</style>\n", fp );
  }
  fprintf( fp, "</head>\n<body>\n<h1>%s</h1>\n", title );
}


void htmlfooters( FILE* fp )
{
  fputs( "</body>\n</html>\n", fp );
}


/**
 * See usage() for usage information.
 */
int main( int argc, char* argv[] )
{
  typedef enum { UNDEFINED, EMBED, EXT, PRE, CSS, ANSI, ANSIUTF, ANSIBOM } Mode;
  Mode	 mode = UNDEFINED;
  char*  outfn = NULL, *cssfn = CSSNAME, *title = NULL;
  LPWSTR cdump;
  LPWORD adump;
  WCHAR  titleW[MAX_PATH*2];
  char	 title8[MAX_PATH*2*4];
  int	 c;
  FILE*  outfp = NULL;
  char*  attribs_used = NULL;
  BOOL	 direct = FALSE;
  BOOL	 key = FALSE;

  if (argc > 1)
  {
    if (strcmp( argv[1], "--help" ) == 0)
      usage();
    if (strcmp( argv[1], "--version" ) == 0)
    {
      puts( "Console Dump version " PVERS " (" PDATE ")." );
      exit( 0 );
    }
  }
  
  // parse arguments
  while ((c = getopt( argc, argv, "abCdehkmpuUw?c:l:o:s:t:" )) > 0)
  {
    switch (c)
    {
      case 'h': mode = EMBED;   break;
      case 'e': mode = EXT;     break;
      case 'p': mode = PRE;     break;
      case 'C': mode = CSS;     break;
      case 'a': mode = ANSI;    break;
      case 'u': mode = ANSIUTF; break;
      case 'U': mode = ANSIBOM; break;

      case 'd': direct = TRUE;  break;
      case 'k': key    = TRUE;  break;
      case 'w': window = TRUE;  break;
      case 'b': lines = 32767;  break;
      case 's': skip  = atoi( optarg ); break;
      case 'l': lines = atoi( optarg ); break;
      case 't': title = optarg; break;
      case 'c': cssfn = optarg; break;
      case 'o': outfn = optarg; break;

      default: usage();
    }
  }

  // any leftover args would be erroneus
  if (optind < argc || mode == UNDEFINED)
    usage();

  // if dumping CSS only, dump CSS, and exit
  if (mode == CSS)
  {
    if (!strcmp( cssfn, "-" ))
      outfp = stdout;
    else
    {
      if (!(outfp = fopen( cssfn, "wb" )))
      {
	perror( cssfn );
	exit( 3 );
      }
    }
    dumpcss( outfp, 0 );
    fclose( outfp );
    exit( 0 );
  }

  if (key)
  {
    while (_kbhit()) _getch();
    while (!_kbhit()) Sleep( 0 );
    if (_getch() == 27)
      exit( 1 );
    while (_kbhit()) _getch();
  }

  if (readcon( &cdump, &adump ))
    exit( 2 );

  // get the console's title, if necessary
  if (title == 0 && mode < PRE)
  {
    LPWSTR dash;
    if (GetConsoleTitleW( titleW, MAX_PATH*2 ) == 0)
      strcpy( title8, "Console" );
    else
    {
      // strip out our name - "Command Prompt - condump..." -> "Command Prompt"
      dash = wcsstr( titleW, L" - " );
      if (dash)
	*dash = '\0';
      WideCharToMultiByte( CP_UTF8, 0, titleW, -1, title8, sizeof(title8),
			   NULL, NULL );
    }
    title = title8;
  }

  // open output file (or point to stdout)
  if (outfn == NULL)
    outfn = (mode < ANSI) ? HTMLNAME : ANSINAME;
  if (strcmp( outfn, "-" ))
  {
    if (!(outfp = fopen( outfn, "wb" )))
    {
      perror( outfn );
      exit( 3 );
    }
  }
  else
    outfp = stdout;

  // encode dump as requested
  switch (mode)
  {
    default: // quiet gcc about missing enumeration values
    case EMBED: // embed stylesheet in HTML <HEAD>
      if (!direct)
      {
	if (!(attribs_used = calloc( 256, 1 )))
	{
	  fputs( "Can't allocate memory for attributes used\n", stderr );
	  exit( 4 );
	}
	scanattribs( adump, attribs_used );
      }
      htmlheaders( outfp, title, NULL, attribs_used );
      free( attribs_used );
      htmlencode( outfp, cdump, adump, direct );
      htmlfooters( outfp );
      break;

    case EXT: // refer to external CSS
      htmlheaders( outfp, title, (direct) ? NULL : cssfn, NULL );
      htmlencode( outfp, cdump, adump, direct );
      htmlfooters( outfp );
      break;

    case PRE: // <pre>...</pre> fragment only
      htmlencode( outfp, cdump, adump, direct );
      break;

    case ANSI:
      ansiencode( outfp, cdump, adump, GetConsoleOutputCP(), direct );
      break;

    case ANSIBOM:
      fputs( "\xEF\xBB\xBF", outfp );
    case ANSIUTF:
      ansiencode( outfp, cdump, adump, CP_UTF8, direct );
      break;
  }

  // all done
  fclose( outfp );
  return 0;
}
